

from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import relationship

from tendril.utils.db import DeclBase
from tendril.utils.db import BaseMixin
from tendril.utils.db import TimestampMixin

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class Provider(DeclBase, BaseMixin):
    name = Column(String(50), unique=True)
    users = relationship("User", back_populates="provider")


class UserRoles(DeclBase, BaseMixin, TimestampMixin):
    user_id = Column(Integer(),
                     ForeignKey('User.id', ondelete='CASCADE'))
    role_id = Column(Integer(),
                     ForeignKey('Role.id', ondelete='CASCADE'))


class User(DeclBase, BaseMixin, TimestampMixin):
    puid = Column(String(255), nullable=False)

    roles = relationship('Role', secondary=UserRoles.__table__,
                         back_populates='users')

    provider_id = Column(Integer(),
                         ForeignKey('Provider.id'), nullable=False)
    provider = relationship("Provider", back_populates="users")

    __table_args__ = (
        UniqueConstraint('puid', 'provider_id'),
    )


class Role(DeclBase, BaseMixin):
    name = Column(String(50), unique=True)
    description = Column(String(255))
    users = relationship('User', secondary=UserRoles.__table__,
                         back_populates='roles')






from pydantic import Field
from pydantic import HttpUrl
from pydantic import validator
from pydantic import create_model

from tendril.utils.pydantic import TendrilTBaseModel


class UserStubTModel(TendrilTBaseModel):
    name: str = Field(..., example="User Full Name")
    nickname: str = Field(..., example="nickname")
    picture: HttpUrl = Field(..., example='https://s.gravatar.com/avatar/...')
    user_id: str = Field(..., example='auth0|...')


def _expand_user_stub(*args, **kwargs):
    # This strangeness is here to (hopefully)
    # break a circular import issue
    from tendril.authn.users import expand_user_stub
    return expand_user_stub(*args, **kwargs)


def UserStubTMixin(inp='puid', out='user'):
    validators = {
        'expand_user_stub':
        validator('puid', pre=True)(_expand_user_stub)
    }
    kwargs = {
        inp : (UserStubTModel, Field(..., alias=out)),
        '__base__': TendrilTBaseModel,
        '__validators__': validators
    }
    _inner = create_model(
        f'UserStubTModel_{inp}_{out}',
        **kwargs
    )
    return _inner





from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr


class UserMixin(object):
    @declared_attr
    def user_id(cls):
        return Column(Integer(), ForeignKey('User.id'), nullable=False)

    @declared_attr
    def user(cls):
        return relationship("User")





import importlib

from tendril.config import AUTH0_PATCH_SCOPES_ON_STARTUP
from tendril.utils.versions import get_namespace_package_names
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class ScopesManager(object):
    def __init__(self, prefix):
        self._prefix = prefix
        self._scopes = {}
        self._default_scopes = []
        self._find_scopes()
        self.finalized = False

    def _create_scopes(self):
        from tendril.authz import connector
        connector.commit_scopes(self._scopes)

    def _find_scopes(self):
        logger.debug("Loading authn scopes from {0}".format(self._prefix))
        modules = list(get_namespace_package_names(self._prefix))
        for m_name in modules:
            if m_name == __name__:
                continue
            m = importlib.import_module(m_name)
            logger.debug("Loading scopes from {0}".format(m_name))
            self._scopes.update(m.scopes)
            if hasattr(m, 'default_scopes'):
                self._default_scopes.extend(m.default_scopes)

    def finalize(self):
        self.finalized = True
        if AUTH0_PATCH_SCOPES_ON_STARTUP:
            self._create_scopes()

    @property
    def scopes(self):
        return self._scopes

    @property
    def default_user_scopes(self):
        return self._default_scopes





from tendril.config import AUTH_PROVIDER
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


if AUTH_PROVIDER == "auth0":
    logger.info("Using the auth0 authz provider")
    from tendril.authz import auth0 as AuthProvider
    provider_name = 'auth0'
else:
    raise ImportError("AUTH_PROVIDER {} not recognized".format(AUTH_PROVIDER))


def get_current_scopes():
    return AuthProvider.get_current_scopes()


def commit_scopes(scopes):
    return AuthProvider.commit_scopes(scopes)


def get_user_scopes(user_id):
    return AuthProvider.get_user_scopes(user_id)


def add_user_scopes(user_id, scopes):
    return AuthProvider.add_user_scopes(user_id, scopes)


def remove_user_scopes(user_id, scopes):
    return AuthProvider.remove_user_scopes(user_id, scopes)





import os

from tendril.config import AUTH0_NAMESPACE
os.environ['AUTH0_RULE_NAMESPACE'] = AUTH0_NAMESPACE

from fastapi import Security
from fastapi_auth0 import Auth0, Auth0User

from auth0.v3.exceptions import Auth0Error
from auth0.v3.authentication import GetToken
from auth0.v3.management import Auth0 as Auth0PythonClient

from tendril.authz import scopes

from tendril.config import AUTH0_DOMAIN
from tendril.config import AUTH0_AUDIENCE
from tendril.config import AUTH0_USER_MANAGEMENT_API_CLIENTID
from tendril.config import AUTH0_USER_MANAGEMENT_API_CLIENTSECRET
from tendril.config import AUTH0_USERINFO_CACHING
from tendril.config import AUTH0_M2M_CLIENTS


from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)


logger.info("Using Auth0 parameters:\n"
            "  - domain    {} \n"
            "  - audience  {} \n"
            "  - namespace {} ".format(AUTH0_DOMAIN, AUTH0_AUDIENCE, AUTH0_NAMESPACE))


if AUTH0_USERINFO_CACHING == 'platform':
    logger.info("Using platform level caching for Auth0 UserInfo")
    from tendril.caching import platform_cache as cache
else:
    logger.info("Not caching Auth0 UserInfo")
    from tendril.caching import no_cache as cache


auth = Auth0(domain=AUTH0_DOMAIN,
             api_audience=AUTH0_AUDIENCE,
             scopes=scopes.scopes)


authn_dependency = auth.implicit_scheme
AuthUserModel = Auth0User

management_api_token = None


def auth_spec(scopes=None):
    kwargs = {}
    if scopes:
        kwargs['scopes'] = scopes
    return Security(auth.get_user, **kwargs)


def get_management_api_token():
    global management_api_token
    logger.debug("Attempting to get the management API token using:\n"
                 "  - domain          {}\n"
                 "  - client id       {}\n"
                 "  - client secret   ending with {}".format(
        AUTH0_DOMAIN, AUTH0_USER_MANAGEMENT_API_CLIENTID,
        AUTH0_USER_MANAGEMENT_API_CLIENTSECRET[-5:0])
    )
    get_token = GetToken(AUTH0_DOMAIN)
    token = get_token.client_credentials(AUTH0_USER_MANAGEMENT_API_CLIENTID,
                                         AUTH0_USER_MANAGEMENT_API_CLIENTSECRET,
                                         'https://{}/api/v2/'.format(AUTH0_DOMAIN))
    management_api_token = token['access_token']
    logger.info("Successfully received Management API token ending in {}".format(management_api_token[-5:]))


def management_api(func):
    def _wrapper(*args, **kwargs):
        global management_api_token
        if management_api_token is None:
            get_management_api_token()
        try:
            auth0 = Auth0PythonClient(AUTH0_DOMAIN, management_api_token)
            return func(*args, auth0=auth0, **kwargs)
        except Auth0Error as error:
            if error.status_code == 401:
                get_management_api_token()
                auth0 = Auth0PythonClient(AUTH0_DOMAIN, management_api_token)
                return func(*args, auth0=auth0, **kwargs)
            if error.status_code == 400:
                # this may be an M2M client
                raise
            else:
                raise
    return _wrapper


@management_api
def get_user_object(user_id, auth0=None):
    logger.debug("Attempting to fetch user information for {} from Auth0".format(user_id))
    user_profile = auth0.users.get(user_id)
    logger.info("Got user details for {} from Auth0 Management API".format(user_id))
    return user_profile


_m2m_clients = {}


def _key_func(user_id):
    return user_id


@cache(namespace='userinfo', ttl=86400, key=_key_func)
def get_user_profile(user_id):
    if user_id in _m2m_clients.keys():
        return _m2m_clients[user_id]
    global management_api_token
    if management_api_token is None:
        get_management_api_token()
    try:
        return get_user_object(user_id)
    except Auth0Error as error:
        if error.status_code == 401:
            get_management_api_token()
            return get_user_object(user_id)
        if error.status_code == 400:
            # this may be an M2M client
            raise
        else:
            raise


@cache(namespace='userstub', ttl=86400, key=_key_func)
def get_user_stub(user_id):
    profile = get_user_profile(user_id)
    return {
        'name': profile['name'],
        'nickname': profile['nickname'],
        'picture': profile['picture'],
        'user_id': profile['user_id'],
    }


def init():
    global _m2m_clients
    _m2m_clients = AUTH0_M2M_CLIENTS





from sqlalchemy.orm.exc import NoResultFound
from tendril.utils.db import with_db

from .model import User
from .model import Provider

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


@with_db
def get_users(session=None):
    pass


@with_db
def get_user(puid, provider=None, session=None):
    filters = [User.puid == puid]
    if provider:
        filters.append(User.provider == provider)
    q = session.query(User).filter_by(*filters)
    return q.one()


@with_db
def get_user_by_id(id, session=None):
    return session.query(User).get(id)


@with_db
def get_provider(name, session=None):
    q = session.query(Provider).filter_by(name=name)
    return q.one()


@with_db
def register_provider(name, must_create=False, session=None):
    if name is None:
        raise AttributeError("name cannot be None")

    try:
        existing = get_provider(name)
    except NoResultFound:
        provider = Provider(name=name)
    else:
        if must_create:
            raise ValueError("Provider Already Exists")
        else:
            provider = existing
    session.add(provider)
    return provider


@with_db
def register_user(puid, provider, must_create=False, session=None):
    if puid is None:
        raise AttributeError("puid cannot be None")
    if provider is None:
        raise AttributeError("provider cannot be None")

    if not isinstance(provider, Provider):
        provider = get_provider(name=provider, session=session)

    try:
        q = session.query(User).filter_by(
            puid=puid, provider=provider
        )
        existing = q.one()
    except NoResultFound:
        logger.info("Registering user '{}' from provider '{}'"
                    "".format(puid, provider.name))
        first_login = True
        user = User(puid=puid, provider=provider, provider_id=provider.id)
    else:
        if must_create:
            raise ValueError("User Already Exists")
        else:
            logger.info("Using existing user '{}' for {} user '{}'"
                        "".format(existing.id, provider.name, puid))
            user = existing
            first_login = False
    session.add(user)
    return user, first_login


@with_db
def preprocess_user(user, provider=None, session=None):
    if user is None:
        raise AttributeError("user cannot be None")
    if isinstance(user, int):
        user_id = user
    elif isinstance(user, User):
        user_id = user.id
    else:
        try:
            user = get_user(user, provider, session=session)
            user_id = user.id
        except NoResultFound:
            raise AttributeError(f"User {user} does not seem to exist.")
    return user_id





from fastapi import APIRouter
from fastapi import Depends

from tendril.authn.users import authn_dependency
from tendril.authn.users import AuthUserModel
from tendril.authn.users import auth_spec
from tendril.authn.users import verify_user_registration
from tendril.authn.users import get_user_profile
from tendril.authn.users import get_user_stub


user_services = APIRouter(prefix='/user',
                          tags=["User Authentication Services"],
                          dependencies=[Depends(authn_dependency),
                                        auth_spec()])


@user_services.get("/verify")
async def verify(user: AuthUserModel = auth_spec()):
    verify_user_registration(user)
    return {"message": "Logged in User Verified."}


@user_services.get("/profile/me")
async def my_profile(user: AuthUserModel = auth_spec()):
    return get_user_profile(user)


@user_services.get("/stub")
async def user_stub(user_id: str):
    # TODO change this to use the nickname or something instead
    return get_user_stub(user_id)


routers = [
    user_services
]





from tendril.config import AUTH_PROVIDER
from tendril.utils.pydantic import TendrilTBaseModel
from tendril.authn.pydantic import UserStubTModel

from .db.model import User
from .db.controller import register_provider
from .db.controller import register_user
from .db.controller import get_user_by_id


from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)


if AUTH_PROVIDER == "auth0":
    logger.info("Using the auth0 auth provider")
    from . import auth0 as AuthProvider
    provider_name = 'auth0'
else:
    raise ImportError("AUTH_PROVIDER {} not recognized".format(AUTH_PROVIDER))


authn_dependency = AuthProvider.authn_dependency
AuthUserModel = AuthProvider.AuthUserModel
auth_spec = AuthProvider.auth_spec
get_provider_user_profile = AuthProvider.get_user_profile


def preprocess_user(user):
    if isinstance(user, AuthUserModel):
        user = user.id
    elif isinstance(user, User):
        user = user.id
    elif isinstance(user, int):
        user = get_user_by_id(user).puid
    return user


def get_user_profile(user):
    user = preprocess_user(user)
    profile = {}
    profile[provider_name] = get_provider_user_profile(user)
    return profile


def expand_user_stub(cls, v):
    if isinstance(v, str):
        return get_user_stub(v)
    return v


def get_user_stub(user):
    user = preprocess_user(user)
    return AuthProvider.get_user_stub(user)


def verify_user_registration(user):
    user_id = preprocess_user(user)
    user, first_login = register_user(user_id, provider_name)
    if first_login:
        from tendril.authz.connector import add_user_scopes
        from tendril.authz.scopes import default_user_scopes
        add_user_scopes(user.puid, default_user_scopes)


def init():
    AuthProvider.init()
    register_provider(provider_name)


init()





from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']


config_elements_auth = [
    ConfigOption(
        'AUTH_PROVIDER',
        "'auth0'",
        "Auth Provider"
    ),
]


config_elements_auth0 = [
    ConfigOption(
        'AUTH0_DOMAIN',
        "None",
        "Auth0 Domain"
    ),
    ConfigOption(
        'AUTH0_AUDIENCE',
        "None",
        "Auth0 Audience"
    ),
    ConfigOption(
        'AUTH0_AUDIENCE_ID',
        'None',
        'Auth0 Resource ID for the Auth0 Audience '
        'provided by the application.'
    ),
    ConfigOption(
        'AUTH0_NAMESPACE',
        "'https://tendril.link/schema/auth0'",
        "Auth0 Namespace for Token Contents"
    ),
    ConfigOption(
        'AUTH0_USER_MANAGEMENT_API_CLIENTID',
        "None",
        "Client ID for interaction with the Auth0 Management API. "
        "Required for User database integration."
    ),
    ConfigOption(
        'AUTH0_USER_MANAGEMENT_API_CLIENTSECRET',
        "None",
        "Client Secret for interaction with the Auth0 Management API. "
        "Required for User database integration."
    ),
    ConfigOption(
        'AUTH0_USERINFO_CACHING',
        "None",
        "Whether to cache userinfo acquired from the management API. "
        "Set to 'platform' for using platform-level caching, using redis. "
        "Other options not presently implemented."
    ),
    ConfigOption(
        'AUTH0_M2M_CLIENTS',
        '{}',
        "A dictionary of M2M clients which are allowed to access this instance. "
        "This needs to be specified here to prevent an attempt to get the "
        "User Profile from Auth0, which would fail. Use of the management "
        "API along with exception handling would be a better approach "
        "to do this."
    ),
    ConfigOption(
        'AUTH0_PATCH_SCOPES_ON_STARTUP',
        'False',
        "Whether to patch scopes on Auth0 at startup."
    )
]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_auth,
                          doc="Authentication Configuration")
    if manager.AUTH_PROVIDER == "auth0":
        manager.load_elements(config_elements_auth0,
                              doc="Auth0 Configuration")






from tendril.authn.auth0 import Auth0PythonClient
from tendril.authn.auth0 import management_api
from tendril.config import AUTH0_AUDIENCE_ID
from tendril.config import AUTH0_AUDIENCE


from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)


@management_api
def get_resource_servers(auth0: Auth0PythonClient):
    return auth0.resource_servers


def get_resource_server():
    return get_resource_servers().get(AUTH0_AUDIENCE_ID)


def get_current_scopes():
    return {x['value']: x['description'] for x in get_resource_server()['scopes']}


def commit_scopes(scopes):
    logger.debug("Patching scopes on Auth0")
    rs = get_resource_servers()
    rs.update(AUTH0_AUDIENCE_ID,
              {'scopes': [{'value': k, 'description': v}
                          for k, v in scopes.items()]})


@management_api
def get_user_scopes(user_id, auth0: Auth0PythonClient):
    return [x['permission_name'] for x in
            auth0.users.list_permissions(user_id, per_page=100)['permissions']]


@management_api
def add_user_scopes(user_id, scopes, auth0: Auth0PythonClient):
    return auth0.users.add_permissions(
        user_id,
        [{"resource_server_identifier": AUTH0_AUDIENCE,
          "permission_name": x} for x in scopes])


@management_api
def remove_user_scopes(user_id, scopes, auth0: Auth0PythonClient):
    return auth0.users.remove_permissions(
        user_id,
        [{"resource_server_identifier": AUTH0_AUDIENCE,
          "permission_name": x} for x in scopes])
